export default class {

  constructor(width, height, tileSize = 4, aliveColor = "white", deadColor = "black", gridColor = "#252525") {
    this.width = width;
    this.height = height;
    this.tileSize = tileSize;
    this.aliveColor = aliveColor;
    this.deadColor = deadColor;
    this.gridColor = gridColor;
    this.map = [];
  }

  randomize(isAliveProb = 0.5) {
    for (let x = 0; x < this.width; x++) {
      this.map[x] = [];
      for (let y = 0; y < this.height; y++) {
        this.map[x][y] = Math.random() < isAliveProb;
      }
    }
  }

  makeBorder(isAlive = false) {
    for (let x = 0; x < this.width; x++) {
      this.map[x][0] = false;
      this.map[x][this.height - 1] = isAlive;
    }
    for (let y = 0; y < this.height; y++) {
      this.map[0][y] = false;
      this.map[this.width - 1][y] = isAlive;
    }
  }

  createSquare(posX, posY, size, isAlive = true) {
    let startX = Math.max(0, posX);
    let endX = Math.min(this.width - 1, posX + size);
    let startY = Math.max(0, posY);
    let endY = Math.min(this.height - 1, posY + size);
    for (let x = startX; x <= endX; x++) {
      for (let y = startY; y <= endY; y++) {
        this.map[x][y] = isAlive;
      }
    }
  }

  resize(factor) {
    let newMap = [];
    this.width *= factor;
    this.height *= factor;
    for (let x = 0; x < this.width; x++) {
      newMap[x] = [];
      for (let y = 0; y < this.height; y++) {
        newMap[x][y] = this.map[Math.floor(x / factor)][Math.floor(y / factor)];
      }
    }
    this.tileSize = Math.floor(this.tileSize / factor);
    this.map = newMap;
  }

  toggleState(x, y, ctx = null) {
    this.map[x][y] = !this.map[x][y];
    if (ctx) {
      this.drawTile(x, y, this.map[x][y], ctx);
    }
  }

  applyRuleNtimes(birthRule, survivalRule, ntimes = 1, ctx = null) {
    for (let i = 0; i < ntimes; i++) {
      this.applyRule(birthRule, survivalRule, ctx);
    }
  }

  applyRule(birthRule, survivalRule, ctx = null) {
    let newMap = [];
    for (let x = 0; x < this.map.length; x++) {
      newMap[x] = [];
      for (let y = 0; y < this.map[x].length; y++) {
        if (this.map[x][y]) {
          newMap[x][y] = survivalRule[this.countAliveMooreNeighborhood(x, y) - 1];
        } else {
          newMap[x][y] = birthRule[this.countAliveMooreNeighborhood(x, y)];
        }
        //draw change if ctx is given
        if (ctx && newMap[x][y] != this.map[x][y]) {
          this.drawTile(x, y, newMap[x][y], ctx);
        }
      }
    }
    this.map = newMap;
  }

  countAliveMooreNeighborhood(posX, posY, chebyshevDistance = 1) {
    let startX = Math.max(0, posX - chebyshevDistance);
    let endX = Math.min(this.width - 1, posX + chebyshevDistance);
    let startY = Math.max(0, posY - chebyshevDistance);
    let endY = Math.min(this.height - 1, posY + chebyshevDistance);
    let nbAliveNeighbors = 0;
    for (let x = startX; x <= endX; x++) {
      for (let y = startY; y <= endY; y++) {
        if (this.map[x][y]) {
          nbAliveNeighbors++;
        }
      }
    }
    return nbAliveNeighbors;
  }

  drawGrid(ctx) {
    ctx.strokeStyle = this.gridColor;
    for (let x = 0; x < this.map.length; x++) {
      for (let y = 0; y < this.map[x].length; y++) {
        ctx.strokeRect(x * this.tileSize, y * this.tileSize, this.tileSize, this.tileSize);
      }
    }
  }

  drawTile(x, y, state, ctx) {
    ctx.fillStyle = state ? this.aliveColor : this.deadColor;
    ctx.fillRect(x * this.tileSize + 1, y * this.tileSize + 1, this.tileSize - 1, this.tileSize - 1);
  }

  drawTiles(ctx) {
    for (let x = 0; x < this.map.length; x++) {
      for (let y = 0; y < this.map[x].length; y++) {
        this.drawTile(x, y, this.map[x][y], ctx);
      }
    }
  }

  draw(ctx) {
    this.drawGrid(ctx);
    this.drawTiles(ctx);
  }

}