import LikeLifeAutomaton from "class/LifeLikeAutomaton";

export default class extends LikeLifeAutomaton {

  getFlowTargetAt(x, y) {
    if (!this.isValidPos(x,y) || this.flowMap[x][y] === false) {
      return false;
    }
    return {...this.flowMap[x][y]};
  }

  getFlowPathFrom(posX, posY) {
    let path = [];
    let x = posX;
    let y = posY;
    let target = this.getFlowTargetAt(x, y);
    while (target !== false) {
      path.push(target);
      x = target.x;
      y = target.y;
      target = this.getFlowTargetAt(x, y);
    }
    return path;
  }

  getFlowAngleAt(x, y) {
    let flowTo = this.getFlowTargetAt(x, y);
    if (flowTo === false) return false;
    let angle = Math.atan2(flowTo.y - y, flowTo.x - x);
    if (angle < 0 ) {
      angle = angle + 2 * Math.PI;
    }
    return angle;
  }

  getNewFlowMap() {
    let flowMap = [];
    for (let x = 0; x < this.width; x++) {
      flowMap[x] = [];
      for (let y = 0; y < this.height; y++) {
        flowMap[x][y] = false;
      }
    }
    return flowMap;
  }

  isReachable(from, to) {
    return true;
  }

  isValidPos(x, y) {
    return x >= 0 && x < this.width
        && y >= 0 && y < this.height;
  }

  isWalkable(x, y) {
    return this.isValidPos(x, y) && this.map[x][y];
  }

  getWalkableNeighbors(x, y, withDiag = false, limitedDiagMove = true) {
    // Von Neumann neighborhood (without itself)
    let neighbors = [
      {x: x + 1, y: y},
      {x: x, y: y + 1},
      {x: x - 1, y: y},
      {x: x, y: y - 1}
    ];
    // asthetic only, not needed
    if ((x + y) % 2 == 0) {
      neighbors.reverse();
    }
    // Moore neighborhood (without itself) if needed
    if (withDiag) {
      let diagNeighbors = [
        {x: x + 1, y: y - 1},
        {x: x + 1, y: y + 1},
        {x: x - 1, y: y + 1},
        {x: x - 1, y: y - 1}
      ];
      if (limitedDiagMove) {
        // Disallow diag move, if the two adjacents neighbors are not walkable
        diagNeighbors = diagNeighbors.filter(pos =>
          this.isWalkable(pos.x + (x - pos.x), pos.y)
          && this.isWalkable(pos.x, pos.y + (y - pos.y))
        );
      }
      neighbors = neighbors.concat(diagNeighbors);
    }
    neighbors = neighbors.filter(pos => this.isWalkable(pos.x, pos.y));
    return neighbors;
  }

  // breadthFirstSearch flow map maker
  flowFieldTo(x, y, allowDiag = true, modifiyInternalFlow = true) {
    let flowMap = this.getNewFlowMap();
    let frontier = [];
    frontier.push({x, y});
    flowMap[x][y] = {x, y};
    while (frontier.length > 0) {
      let current = frontier.shift();
      this.getWalkableNeighbors(current.x, current.y, allowDiag).forEach(next => {
        if (flowMap[next.x][next.y] === false) {
          frontier.push(next);
          flowMap[next.x][next.y] = {x: current.x, y: current.y};
        }
      });
    }
    flowMap[x][y] = false;
    if (modifiyInternalFlow) {
      this.flowMap = flowMap;
    }
    return flowMap;
  }

  drawFlowArrows(ctx) {
    ctx.fillStyle = this.gridColor;
    ctx.font = `${this.tileSize/2}px serif`;
    ctx.textAlign = "center";
    ctx.textBaseline = "middle";
    for (let x = 0; x < this.map.length; x++) {
      for (let y = 0; y < this.map[x].length; y++) {
        if (this.flowMap[x][y] !== false) {
          let flowTo = this.flowMap[x][y];
          let angle = Math.atan2(flowTo.y - y, flowTo.x - x);
          let centerX = x * this.tileSize + this.tileSize / 2;
          let centerY = y * this.tileSize + this.tileSize / 2;
          ctx.translate(centerX, centerY);
          ctx.rotate(angle);
          ctx.fillText('→', 0, 0);
          ctx.setTransform(1, 0, 0, 1, 0, 0);
        }
      }
    }
  }

}