import Keyboard from "class/Keyboard";
import Td from "class/LifeLikeAutomaton/WithPathfinding";
import Mob from "class/Circle/WithEyes";
import randomColor from "randomcolor";
import TWEEN from '@tweenjs/tween.js';

const KEYBOARD = new Keyboard();

let td = {
  ctx: $("canvas").get(0).getContext("2d"),
  grid: null,
  allowDiagMove: true,
  tileSize: 40,
  mobs: new Map(),
  lastTick: null,
  mobSpawnTimer: 0,
  spawnMobEvery: 500,
  mobSpeed: 0.1,
  flowTo: {x: 10, y: 10},
  flowFrom: {x: 0, y: 0},
  animationLoop: null,
  elapsedTime: 0,
  pauseAutoTrigged: false,
  birthRule_cave: [false, false, false, false, false, true, true, true, true],
  survivalRule_cave: [false, false, false, true, true, true, true, true, true],

  tick(now) {
    this.animationLoop = requestAnimationFrame(now => this.tick(now));
    let deltaT = now - this.lastTick;
    this.lastTick = now;
    //if less than 30fps (Δt > ~33,3 ms) drop the frame
    if (deltaT > 34) return;
    this.elapsedTime += deltaT;
    this.mobSpawnTimer += deltaT;
    if (this.mobSpawnTimer >= this.spawnMobEvery) {
      this.generateMob();
      this.mobSpawnTimer = this.mobSpawnTimer - this.spawnMobEvery;
    }
    TWEEN.update(this.elapsedTime);
    this.ctx.canvas.width = this.ctx.canvas.clientWidth;
    this.ctx.canvas.height = this.ctx.canvas.clientHeight;
    this.grid.draw(this.ctx);
    this.grid.drawFlowArrows(this.ctx);
    this.drawStartEnd();
    this.mobs.forEach(mob => mob.draw(this.ctx));
  },

  drawStartEnd() {
    this.ctx.fillStyle = "green";
    this.ctx.fillRect(this.flowFrom.x * this.tileSize + 1, this.flowFrom.y * this.tileSize + 1, this.tileSize - 1, this.tileSize - 1);
    this.ctx.fillStyle = "red";
    this.ctx.fillRect(this.flowTo.x  * this.tileSize + 1, this.flowTo.y * this.tileSize + 1, this.tileSize - 1, this.tileSize - 1);
  },

  createTweenMob(mob) {
    let x = Math.floor(mob.x / this.tileSize);
    let y = Math.floor(mob.y / this.tileSize);
    let path = this.grid.getFlowPathFrom(x, y);
    // If no path, we are at an inacessible place (and the mob die)
    if (path.length == 0) {
      this.mobs.delete(mob);
      return;
    }
    let lastTween = null;
    let firstTween = null;
    let from = {x: mob.x, y: mob.y};
    let lastAngle = mob.eyesAngle;
    // Chaining tweens from start to dest (waypoints management in between)
    path.forEach(flowTo => {
      let tween = new TWEEN.Tween(mob);
      let angle = this.grid.getFlowAngleAt(flowTo.x, flowTo.y);
      // make the eyes roll less frequent ^_^
      if (Math.abs(lastAngle - angle) > Math.PI){
          angle += Math.sign(lastAngle - angle) * 2 * Math.PI;
      }
      let dest = {
        x: flowTo.x * this.tileSize + this.tileSize / 2,
        y: flowTo.y * this.tileSize + this.tileSize / 2,
        eyesAngle: angle
      };
      // Δtime = Δdist / mob.speed
      // Δdist is calculated with Pythagorean theorem
      let distance = Math.sqrt(Math.pow(from.x - dest.x, 2) + Math.pow(from.y - dest.y, 2));
      tween.to(dest, distance / mob.speed);
      if (firstTween === null) {
        firstTween = tween;
      } else {
        lastTween.chain(tween);
      }
      lastAngle = angle;
      from = dest;
      lastTween = tween;
    });
    lastTween.onComplete(mob => {
      // The last tween complete. We are at our destination (and the mob vanish)
      this.mobs.delete(mob);
    });
    firstTween.start(this.elapsedTime);
  },

  generateMob() {
    let mob = new Mob(
      (this.flowFrom.x * this.tileSize) + this.tileSize / 2,
      (this.flowFrom.y * this.tileSize) + this.tileSize / 2,
      this.tileSize / 4,
      this.mobSpeed,
      randomColor(),
      "black",
      this.grid.getFlowAngleAt(this.flowFrom.x, this.flowFrom.y)
    );
    this.mobs.set(mob, mob);
    this.createTweenMob(mob);
  },

  clickHandler(x, y) {
    // disallow click during animation pause
    if (this.animationLoop === null) return;
    let flowFrom = this.flowFrom;
    let flowTo = this.flowTo;
    let toggle = false;
    if (KEYBOARD.isKeyDown('shift')) {
      if (this.grid.isWalkable(x,y)) {
        flowTo = {x, y};
      }
    } else if (KEYBOARD.isKeyDown('control')) {
      if (this.grid.isWalkable(x,y)) {
        flowFrom = {x, y};
      }
    } else {
      toggle = true;
      this.grid.toggleState(x, y);
    }
    // Disallow the modification if destination become inaccessible
    let flowMap = this.grid.flowFieldTo(flowTo.x, flowTo.y, this.allowDiagMove, false);
    if (flowMap[flowFrom.x][flowFrom.y] === false) {
      if (toggle) this.grid.toggleState(x, y);
      return;
    }
    this.flowFrom = flowFrom;
    this.flowTo = flowTo;
    this.grid.flowMap = flowMap;
    // Update all tween to follow the new path
    TWEEN.removeAll();
    this.mobs.forEach(mob => this.createTweenMob(mob));
  },

  startLoop() {
    this.animationLoop = requestAnimationFrame(now => {
      this.lastTick = now;
      this.tick(now);
    });
  },

  stopLoop() {
    cancelAnimationFrame(this.animationLoop);
    this.animationLoop = null;
  },

  toggleLoop() {
    if (this.animationLoop) {
      this.stopLoop();
    } else {
      this.startLoop();
    }
  },

  generateStage() {
    this.grid.randomize(0.52);
    this.grid.createSquare(this.flowFrom.x, this.flowFrom.y, 2);
    this.grid.createSquare(this.flowTo.x, this.flowTo.y, 2);
    this.grid.applyRuleNtimes(this.birthRule_cave, this.survivalRule_cave, 3);
    this.grid.flowFieldTo(this.flowTo.x, this.flowTo.y, this.allowDiagMove);
    if (this.grid.flowMap[this.flowFrom.x][this.flowFrom.y] === false) {
      this.generateStage();
    }
  },

  init() {
    this.ctx.canvas.width = this.ctx.canvas.clientWidth;
    this.ctx.canvas.height = this.ctx.canvas.clientHeight;
    let width = Math.round(this.ctx.canvas.width / this.tileSize) ;
    let height = Math.round(this.ctx.canvas.height / this.tileSize);
    this.grid = new Td(width, height, this.tileSize);
    this.flowTo.x = width - 1;
    this.flowTo.y = height - 1;
    this.generateStage();
    this.startLoop();
    KEYBOARD.onKey("p", () => this.toggleLoop());
    $(this.ctx.canvas).on("click", event => {
      let rect = this.ctx.canvas.getBoundingClientRect();
      let x = Math.floor((event.clientX - rect.left) / this.tileSize);
      let y = Math.floor((event.clientY - rect.top) / this.tileSize);
      this.clickHandler(x, y);
    });
    document.onvisibilitychange = () => {
      if (document.hidden && this.animationLoop) {
        this.pauseAutoTrigged = true;
        this.stopLoop();
      } else if (!document.hidden && this.pauseAutoTrigged && this.animationLoop === null) {
        this.startLoop();
        this.pauseAutoTrigged = false;
      }
    };
  }
};

td.init();